
from abc import abstractmethod

from urwid import Edit, connect_signal

from . import PAGE_WIDTH, Field
from ...lib.utils import label
from ...squeal.tables.statistic import StatisticJournalType
from ..tui.widgets import Rating0, Rating1


class EditableField(Field):

    def widget(self):
        widget = self._widget()
        connect_signal(widget, 'change', self.__on_change)
        return widget

    def __on_change(self, widget, value):
        self._log.debug('Setting %s=%r' % (self._journal.statistic_name.name, value))
        self._journal.value = value


class Text(EditableField):

    statistic_journal_type = StatisticJournalType.TEXT

    def __init__(self, log, journal, width=PAGE_WIDTH):
        super().__init__(log, journal, width=width)

    def _format_value(self, value):
        return repr(value)

    def _widget(self):
        return Edit(caption=label('%s: ' % self._journal.statistic_name.name),
                    edit_text=self._journal.value or '')


class Integer(Field):

    statistic_journal_type = StatisticJournalType.INTEGER

    def __init__(self, log, journal, lo=None, hi=None, width=1):
        super().__init__(log, journal, width=width)
        self._lo = lo
        self._hi = hi

    def _format_value(self, value):
        return '%d' % value

    def _widget(self):
        from ..tui.widgets import Integer
        return Integer(caption=label('%s: ' % self._journal.statistic_name.name), state=self._journal.value,
                       minimum=self._lo, maximum=self._hi, units=self._journal.statistic_name.units)


class Float(EditableField):

    statistic_journal_type = StatisticJournalType.FLOAT

    def __init__(self, log, journal, lo=None, hi=None, dp=2, format='%f', width=1):
        super().__init__(log, journal, width=width)
        self._lo = lo
        self._hi = hi
        self._dp = dp
        self._format = format

    def _format_value(self, value):
        return self._format % value

    def _widget(self):
        from ..tui.widgets import Float
        return Float(caption=label('%s: ' % self._journal.statistic_name.name), state=self._journal.value,
                     minimum=self._lo, maximum=self._hi, dp=self._dp, units=self._journal.statistic_name.units)


class Score(EditableField):

    statistic_journal_type = StatisticJournalType.INTEGER

    def __init__(self, log, journal, width=1):
        super().__init__(log, journal, width=width)

    def _format_value(self, value):
        return '%d' % value

    def _widget(self):
        return self._field_cls(caption=label('%s: ' % self._journal.statistic_name.name), state=self._journal.value)

    @property
    @abstractmethod
    def _field_cls(self):
        pass


class Score0(Score):

    _field_cls = Rating0


class Score1(Score):

    _field_cls = Rating1




from collections import namedtuple
from itertools import repeat, zip_longest

from .support import Named
from ...lib.data import WarnDict

TIMESTAMP_GLOBAL_TYPE = 253


class ScaledField(Named):

    def __init__(self, log, name, units, scale, offset, accumulate):
        super().__init__(log, name)
        self._units = units
        self._scale = scale if scale else 1  # treat 0 as 1
        self._offset = offset if offset else 0
        self._accumulate = accumulate

    def _parse_and_scale(self, type, data, count, endian, timestamp,
                         scale=None, offset=None, accumulators=None, n_bits=None, **options):
        if scale is None: scale = self._scale if self._scale else 1
        if offset is None: offset = self._offset
        values = type.parse_type(data, count, endian, timestamp, scale=scale, offset=offset,
                                 name=self.name, accumulators=accumulators, n_bits=n_bits, **options)
        yield self.name, (values, self._units)

    def register_accumulator(self, accumulators):
        if self._accumulate and self.name not in accumulators:
            accumulators[self.name] = None

    def post(self, message, types):
        '''
        Called after message creation.
        '''
        pass


class TypedField(ScaledField):

    def __init__(self, log, name, field_no, units, scale, offset, accumulate, field_type, types):
        super().__init__(log, name, units, scale, offset, accumulate)
        self.number = field_no
        if name in types.overrides:
            self._log.info('Overriding type for %s (not %s)' % (name, field_type))
            field_type = name
        self.type = types.profile_to_type(field_type)

    def parse_field(self, data, count, endian, timestamp, references, message, **options):
        yield from self._parse_and_scale(self.type, data, count, endian, timestamp, **options)


class RowField(TypedField):

    def __init__(self, log, row, types):
        super().__init__(log, row.field_name, row.single_int(log, row.field_no), row.units,
                         row.single_int(log, row.scale), row.single_int(log, row.offset),
                         row.single_int(log, row.accumulate), row.field_type, types)


class DelegateField(ScaledField):

    def parse_field(self, data, count, endian, timestamp, references, message,
                    scale=None, offset=None, **options):
        scale = self._scale if scale is None else scale
        offset = self._offset if offset is None else offset
        delegate = message.profile_to_field(self.name)
        yield from delegate.parse_field(data, count, endian, timestamp, references, message,
                                        scale=scale, offset=offset, **options)

    def size(self, message):
        # this is needed because we delegate above
        delegate = message.profile_to_field(self.name)
        return delegate.type.n_bytes


class Zip:

    def _zip(self, *fields):
        return zip_longest(*(self.__split(field) for field in fields))

    def __split(self, field):
        if field:
            for value in str(field).split(','):
                yield value.strip()
        else:
            yield None


class CompositeField(Zip, TypedField):

    def __init__(self, log, row, types):
        super().__init__(log, row.field_name, row.single_int(log, row.field_no),
                         None, None, None, None, row.field_type, types)
        self.number = row.single_int(log, row.field_no)
        self._components = []
        self.references = []
        for (name, bits, units, scale, offset, accumulate) in \
                self._zip(row.components, row.bits, row.units, row.scale, row.offset, row.accumulate):
            self._components.append((int(bits),
                                     # set scale / offset below because they seem to override delegate
                                     # (see intensity in CSV examples in SDK inside current_activity_type_intensity)
                                     DelegateField(log, name, units,
                                                    1 if scale is None else float(scale),
                                                    0 if offset is None else float(offset),
                                                    None if accumulate is None else int(accumulate))))
            self.references.append(name)

    def register_accumulator(self, accumulators):
        for _, field in self._components:
            field.register_accumulator(accumulators)

    def parse_field(self, data, count, endian, timestamp, references, message,
                    rtn_composite=False, check_bad=True, n_bits=None, **options):
        if check_bad and self.type.is_bad(data, count, endian):
            yield self.name, (None, self._units)
        else:
            if rtn_composite:  # extra shit for CSV comparison
                yield (self.name, (('COMPOSITE',), self._units))
            byteorder = ['little', 'big'][endian]
            bits = int.from_bytes(data, byteorder=byteorder)
            for n_bits, field in self._components:
                # todo - error if larger
                n_bytes = max((n_bits+7) // 8, field.size(message))
                data = (bits & ((1 << n_bits) - 1)).to_bytes(n_bytes, byteorder=byteorder)
                bits >>= n_bits
                yield from field.parse_field(data, 1, endian, timestamp, references, message,
                                             rtn_composite=rtn_composite, check_bad=False, n_bits=n_bits,
                                             **options)


class DynamicField(Zip, RowField):

    def __init__(self, log, row, rows, types):
        super().__init__(log, row, types)
        self.__dynamic_lookup = WarnDict(log, 'No dynamic field for %r')
        self.references = []
        for row in rows.lookahead():
            if row and row.field_name and row.field_no is None:
                for name, value in self._zip(row.ref_name, row.ref_value):
                    # need to use a list (not set) to preserve order
                    # (and consistently disambiguate multiple matches)
                    if name not in self.references:
                        self.references.append(name)
                    self.__dynamic_lookup[(name, value)] = row.field_name
            else:
                break

    def post(self, message, types):
        # fill in values for when mapping is not used
        for (name, value), field in list(self.__dynamic_lookup.items()):
            value = types.profile_to_type(message.profile_to_field(name).type.name).profile_to_internal(value)
            self.__dynamic_lookup[(name, value)] = field

    def parse_field(self, data, count, endian, timestamp, references, message, warn=False, **options):
        for name in self.references:
            if name in references:
                lookup = (name, references[name][0][0])  # drop units and take first value
                if lookup in self.__dynamic_lookup:
                    yield from message.profile_to_field(self.__dynamic_lookup[lookup]).parse_field(
                        data, count, endian, timestamp, references, message, warn=warn, **options)
                    return
        if warn:
            self._log.warning('Could not resolve dynamic field %s' % self.name)
        yield from super().parse_field(data, count, endian, timestamp, references, message, warn=warn, **options)


def MessageField(log, row, rows, types):
    log.debug('Parsing field %s' % row.field_name)
    if row.components:
        return CompositeField(log, row, types)
    else:
        peek = rows.peek()
        if row.field_no and peek and peek.field_name and peek.field_no is None:
            return DynamicField(log, row, rows, types)
        else:
            return RowField(log, row, types)


class Row(namedtuple('BaseRow',
                     'msg_name, field_no, field_name, field_type, array, components, scale, offset, ' +
                     'units, bits, accumulate, ref_name, ref_value, comment, products, example')):

    __slots__ = ()

    def __new__(cls, row):
        return super().__new__(cls, *tuple(row)[0:16])

    def single_int(self, log, value):
        try:
            return None if value is None else int(value)
        except ValueError:
            log.warning('Cannot parse "%s" as a single integer', value)
            return None




import webbrowser as web
from abc import abstractmethod, ABC
from inspect import getsource, getfullargspec
from logging import getLogger
from os import unlink, makedirs
from os.path import join, exists
from re import compile, sub

import nbformat as nb
import nbformat.v4 as nbv
from nbformat.sign import NotebookNotary

from .server import connection_url, notebook_dir, start_local

log = getLogger(__name__)

QUOTES = "'''"
FQUOTES = 'f' + QUOTES
IPYNB = '.ipynb'


class Token(ABC):

    def __init__(self, vars):
        self._vars = vars

    # called when this token is complete
    def post_one(self):
        pass

    # called when all tokens are complete
    def post_all(self, tokens):
        pass

    @abstractmethod
    def to_cell(self):
        raise NotImplementedError()

    @staticmethod
    def to_notebook(tokens):
        for token in tokens:
            token.post_all(tokens)
        notebook = nbv.new_notebook()
        for token in tokens:
            notebook['cells'].append(token.to_cell())
        return notebook


class TextToken(Token):

    def __init__(self, vars, indent):
        super().__init__(vars)
        self._indent = indent
        self._text = ''

    def append(self, line):
        self._text += line[self._indent:]
        self._text += '\n'

    def __bool__(self):
        return bool(self._text)

    @staticmethod
    def _strip(text):
        while text.startswith('\n'):
            text = text[1:]
        while text.endswith('\n'):
            text = text[:-1]
        return text

    def __repr__(self):
        return f'{self.__class__.__name__}({QUOTES}\n{self._text}\n{QUOTES})'

    def to_cell(self):
        return nbv.new_markdown_cell(self._text)


class Text(TextToken):

    def __init__(self, vars, fmt, indent):
        super().__init__(vars, indent)
        self._fmt = fmt

    @staticmethod
    def _strip_right(text):
        # markdown in jupyter splits lines ending in multiple spaces.
        while text and text[-1] == ' ':
            text = text[:-1]
        return text

    def post_one(self):
        self._text = eval('%s%s%s' % (self._fmt, self._text, QUOTES), self._vars)
        self._text = '\n'.join([self._strip_right(line) for line in self._text.splitlines()])
        self._text = self._strip(self._text)

    def post_all(self, tokens):
        self._text = '\n'.join(self._expand_line(line, tokens) for line in self._text.splitlines())

    def _expand_line(self, line, tokens):
        if line == '$contents':
            return '## Contents\n' + self._contents(tokens)
        else:
            return line

    def _contents(self, tokens):
        sections = []
        for token in tokens:
            if isinstance(token, Text):
                for line in token._text.splitlines():
                    if line.startswith('## '):
                        title = line[3:].strip()
                        sections.append(f'* [{title}](#{title.replace(" ", "-")})')
        return '\n'.join(sections)

    @staticmethod
    def parse(vars, lines, f):
        text = Text(vars, f, 4)
        while lines and lines[0].strip() not in (QUOTES, FQUOTES):
            text.append(lines.pop(0))
        if text:
            text.post_one()
            yield text
        if lines:
            yield from Params.parse_text_or_code(vars, lines[1:])


class Code(TextToken):

    def post_one(self):
        self._text = sub(r'output_file\([^)]*\)', 'output_notebook()', self._text)
        self._text = self._strip(self._text)

    def to_cell(self):
        return nbv.new_code_cell(self._text)

    @staticmethod
    def parse(vars, lines):
        code = Code(vars, 4)
        while lines and lines[0].strip() not in (QUOTES, FQUOTES):
            code.append(lines.pop(0))
        if code:
            code.post_one()
            yield code
        if lines:
            yield from Params.parse_text_or_code(vars, lines)


class Import(Code):

    FILTER = ('decorator import template', '@template')

    def post_one(self):
        self._text = '\n'.join(line for line in self._text.splitlines()
                               if all(filter not in line for filter in self.FILTER))
        self._text = self._text.replace('output_file', 'output_notebook')
        super().post_one()

    @staticmethod
    def parse(vars, lines):
        imports = Import(vars, 0)
        while lines:
            if lines[0].startswith('def '):
                if imports:
                    imports.post_one()
                    yield imports
                yield from Params.parse(vars, lines)
                return
            else:
                imports.append(lines.pop(0))


class Params(Code):

    def __init__(self, vars, params):
        super().__init__(vars, 0)
        self._params = [param.strip() for param in params]

    def post_one(self):
        for param in self._params:
            self.append(f'{param} = "{self._vars[param]}"')
        super().post_one()

    @staticmethod
    def parse(vars, lines):
        line = lines.pop(0)
        template = compile(r'def [^(]+\(([^)]*)\):\s*')
        match = template.match(line)
        if match and match.group(1):
            params = Params(vars, Params.split(match.group(1)))
            params.post_one()
            yield params
        elif not match:
            raise Exception(f'Bad template def: {line}')
        yield from Params.parse_text_or_code(vars, lines)

    @staticmethod
    def split(params):
        return [param.split(':')[0].strip() for param in params.split(',')]


    @staticmethod
    def parse_text_or_code(vars, lines):
        while lines and not lines[0].strip():
            lines.pop(0)
        if lines:
            if lines[0].strip() in (QUOTES, FQUOTES):
                yield from Text.parse(vars, lines[1:], lines[0].strip())
            else:
                yield from Code.parse(vars, lines)


class Help(Text):

    def __init__(self):
        super().__init__({}, False, 0)
        self.append('(Select "Run All" from "Cell" menu above to generate output)')


def tokenize(vars, text):
    yield from Import.parse(vars, list(text.splitlines()))


def load_raw(name):
    log.debug(f'Loading template {name}')
    return getsource(getattr(__import__('ch2.uranus.template', fromlist=[name]), name))


def load_tokens(name, vars):
    log.debug(f'Tokenizing {name}')
    return tokenize(vars, load_raw(name))


def load_notebook(name, vars):
    tokens = list(load_tokens(name, vars))
    tokens = [Help()] + tokens
    return Token.to_notebook(tokens)


def create_notebook(template, args, kargs):

    all_args = ' '.join(args)
    if all_args and kargs: all_args += ' '
    all_args += ' '.join(kargs[key] for key in sorted(kargs.keys()))
    all_args = sub(r'\s+', '-', all_args)

    vars = dict(kargs)
    for name, value in zip(getfullargspec(template).args, args):
        vars[name] = value

    template = template.__name__
    base = join(notebook_dir(), template)
    name = all_args + IPYNB
    path = join(base, name)
    makedirs(base, exist_ok=True)

    log.info(f'Creating {template} at {path} with {vars}')
    notebook = load_notebook(template, vars)
    # https://testnb.readthedocs.io/en/latest/security.html
    NotebookNotary().sign(notebook)
    if exists(path):
        log.warning(f'Deleting old version of {path}')
        unlink(path)
    with open(path, 'w') as out:
        log.info(f'Writing {template} to {path}')
        nb.write(notebook, out)
    return join(template, name)


def display_notebook(template, args, kargs, local=True):
    log.debug(f'Displaying {template} with {args}, {kargs}')
    if local:
        start_local()
    name = create_notebook(template, args, kargs)
    jupyter_url = connection_url()
    url = f'{jupyter_url}tree/{name}'
    log.info(f'Displaying {url}')
    web.open(url, autoraise=False)




from .fields import Row, MessageField, TypedField
from .support import Named
from .support import Rows
from ..format.records import LazyRecord
from ...lib.data import WarnDict

'''
A message is a collection of fields, as defined in the Profile document.

During parsing Definition instances are created which associated a Message with a subset of Fields.

Then, when the Token associated with the Definition is read and needs to be parsed, the Token delegates
the work to the Definition, which in turn delegates the work to the Message, which in turn delegates to 
the Field(s).
'''


class Message(Named):

    def __init__(self, log, name, number=None, warn=False):
        super().__init__(log, name)
        self.number = number
        self._profile_to_field = WarnDict(log, 'No field for profile %r') if warn else dict()
        self._number_to_field = WarnDict(log, 'No field for number %r') if warn else dict()

    def _add_field(self, field):
        self._profile_to_field[field.name] = field
        self._number_to_field[field.number] = field

    def profile_to_field(self, name):
        return self._profile_to_field[name]

    def number_to_field(self, value):
        return self._number_to_field[value]

    def _post(self, types):
        for field in self._number_to_field.values():
            field.post(self, types)

    def parse_message(self, data, defn, timestamp, extra=None, **options):
        return LazyRecord(self.name, self.number, defn.identity, timestamp,
                          self.__parse(data, defn, timestamp, extra=extra, **options))

    def __parse(self, data, defn, timestamp, extra=None, **options):
        # this is the generator that lives inside a record and is evaluated on demand
        if extra is None: extra = {}
        references = {}
        for name, value in extra.items():
            if name in defn.references and value[0] is not None:
                references[name] = value
            yield name, value
        for field in defn.fields:
            bytes = data[field.start:field.finish]
            if field.field:
                for name, value in self._parse_field(
                        field.field, bytes, field.count, defn.endian, timestamp, references, self, **options):
                    if name in defn.references and value[0] is not None:
                        references[name] = value
                    yield name, value
            else:
                name = '@%d:%d' % (field.start, field.finish)
                value = (field.base_type.parse_type(bytes, field.count, defn.endian, timestamp), None)
                yield name, value

    def _parse_field(self, field, bytes, count, endian, timestamp, references, message, **options):
        # allow interception for optional field in header
        yield from field.parse_field(bytes, count, endian, timestamp, references, message, **options)


class RowMessage(Message):

    def __init__(self, log, row, rows, types, warn=False):
        number = types.profile_to_type('mesg_num').profile_to_internal(row.msg_name)
        super().__init__(log, row.msg_name, number, warn=warn)
        while rows:
            if not rows.peek().field_name:
                self._post(types)
                break
            self._add_field(MessageField(self._log, next(rows), rows, types))


class Missing(Message):

    def __init__(self, log, number):
        super().__init__(log, 'MESSAGE %d' % number, number)


class Messages:

    def __init__(self, log, sheet, types, warn=False):
        self.__log = log
        self.__profile_to_message = WarnDict(log, 'No message for profile %r') if warn else dict()
        self.__number_to_message = WarnDict(log, 'No message for number %r') if warn else dict()
        rows = Rows(sheet, wrapper=Row)
        for row in rows:
            if row.msg_name and row.msg_name[0].isupper():
                self.__log.debug('Skipping %s' % (row,))
            elif row.msg_name:
                # self.__log.info('Parsing message %s' % row.msg_name)
                self.__add_message(RowMessage(self.__log, row, rows, types, warn=warn))

    def __add_message(self, message):
        self.__profile_to_message[message.name] = message
        try:
            self.__number_to_message[message.number] = message
        except AttributeError:
            pass

    def profile_to_message(self, name):
        return self.__profile_to_message[name]

    def number_to_message(self, number):
        try:
            return self.__number_to_message[number]
        except KeyError:
            message = Missing(self.__log, number)
            self.__number_to_message[number] = message
            return message






from inspect import getfullargspec
from logging import getLogger
from pkgutil import iter_modules
from signal import pause
from time import sleep

from .args import SUB_COMMAND, SERVICE, START, STOP, SHOW, JUPYTER, LIST, PROGNAME, NAME, ARG, STATUS
from ..lib.workers import command_root
from ..squeal import SystemProcess
from ..uranus import template
from ..uranus.server import JupyterServer, start_service, STARTUP_SLEEP, start_remote

log = getLogger(__name__)


def jupyter(args, db):
    '''
## jupyter

    > ch2 jupyter show ...

Show the template in the browser, starting a background Jupyter server if necessary.

    > ch2 jupyter list

List the available templates and their arguments.

    > ch2 jupyter status

Indicate whether the background server is running or not.

    > ch2 jupyter stop

Stop the background server.
    '''
    if args[SUB_COMMAND] == SERVICE:
        service(db)
    elif args[SUB_COMMAND] == START:
        start(db)
    elif args[SUB_COMMAND] == STOP:
        stop(db)
    elif args[SUB_COMMAND] == SHOW:
        show(db, args)
    elif args[SUB_COMMAND] == STATUS:
        status(db)
    elif args[SUB_COMMAND] == LIST:
        print_list()
    else:
        raise Exception(f'Unexpected command {args[SUB_COMMAND]}')


def service(db):
    with db.session_context() as s:
        start_service(s)
        pause()


def start(db):
    with db.session_context() as s:

        def callback():
            ch2 = command_root()
            log_name = 'jupyter-service.log'
            cmd = f'{ch2} -v0 -l {log_name} {JUPYTER} {SERVICE}'
            SystemProcess.run(s, cmd, log_name, JupyterServer)
            sleep(STARTUP_SLEEP)

        start_remote(s, callback)


def stop(db):
    with db.session_context() as s:
        SystemProcess.delete_all(s, JupyterServer)


def status(db):
    with db.session_context() as s:
        if SystemProcess.exists_any(s, JupyterServer):
            print('\n  Service running\n')
        else:
            print('\n  No service running\n')


def templates():
    for importer, modname, ispkg in iter_modules(template.__path__):
        module = getattr(template, modname)
        function = getattr(module, modname)
        argspec = getfullargspec(function._original)
        yield modname, (function, argspec)


def print_list():
    for name, (_, spec) in templates():
        args = ' '.join(spec.args)
        print(f'\n  {name}  {args}')
    print()


def show(db, args):
    start(db)
    name = args[NAME]
    params = args[ARG]
    try:
        fn, spec = dict(templates())[name]
        params = build_params(params, spec)
        fn(*params, local=False)
    except KeyError:
        raise Exception(f'No template called {name} (see {PROGNAME} {JUPYTER} {LIST})')


def build_params(params, spec):
    if len(params) != len(spec.args):
        raise Exception(f'Received {len(params)} args but need {len(spec.args)} values ' +
                        f'(see {PROGNAME} {JUPYTER} {LIST})')
    return [build_param(param, name, spec.annotations.get(name, None)) for param, name in zip(params, spec.args)]


def build_param(param, name, annotation):
    if annotation:
        return annotation(param)
    else:
        return param





import datetime as dt
from logging import getLogger
from re import sub

from urwid import Pile, Text, Columns

from . import Displayer
from ..calculate.segment import SegmentCalculator
from ..names import SEGMENT_TIME, SEGMENT_HEART_RATE
from ...lib.date import local_date_to_time
from ...squeal.tables.activity import ActivityGroup
from ...squeal.tables.segment import SegmentJournal, Segment
from ...squeal.tables.statistic import StatisticJournal, StatisticName
from ...uweird.fields import ReadOnlyField
from ...uweird.fields.summary import summary_columns
from ...uweird.tui.decorators import Indent

log = getLogger(__name__)


def segments_for_activity(s, ajournal):
    return s.query(SegmentJournal). \
        filter(SegmentJournal.activity_journal == ajournal). \
        order_by(SegmentJournal.start).all()


class SegmentDiary(Displayer):

    def _display_date(self, s, f, date):
        tomorrow = local_date_to_time(date + dt.timedelta(days=1))
        today = local_date_to_time(date)
        pile = []
        # todo - rewrite to use above
        for agroup in s.query(ActivityGroup).order_by(ActivityGroup.sort).all():
            segment_pile = []
            for sjournal in s.query(SegmentJournal). \
                    join(Segment). \
                    filter(SegmentJournal.start >= today,
                           SegmentJournal.start < tomorrow,
                           Segment.activity_group == agroup). \
                    order_by(SegmentJournal.start).all():
                columns = self.__fields(s, date, sjournal)
                if columns:
                    segment_pile.append(Columns(columns))
            if segment_pile:
                pile.append(Pile([Text(sjournal.segment.name), Indent(Pile(segment_pile))]))
        if pile:
            yield Pile([Text('Segments'),
                        Indent(Pile(pile))])

    def __fields(self, s, date, sjournal):
        time = self.__field(s, date, sjournal, SEGMENT_TIME)
        hr = self.__field(s, date, sjournal, SEGMENT_HEART_RATE)
        if time or hr:
            return [time if time else Text(''), hr if hr else Text('')]
        else:
            return None

    def __field(self, s, date, sjournal, name):
        sjournal = StatisticJournal.at_date(s, date, name, SegmentCalculator, sjournal.segment,
                                            source_id=sjournal.id)
        if sjournal:
            return ReadOnlyField(log, sjournal, date=date,
                                 format_name=lambda n: sub(r'^Segment ', '', n)).widget()
        else:
            return None

    def _display_schedule(self, s, f, date, schedule=None):
        rows = []
        for agroup in s.query(ActivityGroup).order_by(ActivityGroup.sort).all():
            group_rows = []
            for segment in s.query(Segment).filter(Segment.activity_group == agroup).all():
                segment_rows = list(self.__schedule_fields(s, f, date, segment, schedule))
                if segment_rows:
                    segment_rows = Pile([Text(segment.name),
                                         Indent(Pile(segment_rows))])
                    group_rows.append(segment_rows)
            if group_rows:
                group_rows = Pile([Text(agroup.name),
                                   Indent(Pile(group_rows))])
                rows.append(group_rows)
        if rows:
            yield Pile([Text('Segments'),
                        Indent(Pile(rows))])

    def __schedule_fields(self, s, f, date, segment, schedule):
        names = list(self.__names(s, segment, SEGMENT_TIME, SEGMENT_HEART_RATE))
        yield from summary_columns(log, s, f, date, schedule, names,
                                   format_name=lambda n: sub(r'^Segment ', '', n))

    def __names(self, s, segment, *names):
        for name in names:
            sname = s.query(StatisticName). \
                filter(StatisticName.name == name,
                       StatisticName.constraint == segment,
                       StatisticName.owner == SegmentCalculator).one_or_none()
            if sname:
                yield sname




from logging import getLogger
from time import time, mktime

from psutil import Popen, pid_exists, Process
from sqlalchemy import Column, Text, Integer

from ..support import Base
from ..types import Time, ShortCls

log = getLogger(__name__)


class SystemConstant(Base):

    __tablename__ = 'system_constant'

    name = Column(Text, primary_key=True)
    value = Column(Text, nullable=False)

    @classmethod
    def get(cls, s, name, none=False):
        instance = s.query(SystemConstant).filter(SystemConstant.name == name).one_or_none()
        if instance is None:
            if not none:
                raise Exception(f'No value for {name}')
            else:
                log.debug(f'Read {name}=None')
                return None
        else:
            log.debug(f'Read {name}={instance.value}')
            return instance.value

    @classmethod
    def set(cls, s, name, value, force=False):
        if force:
            cls.delete(s, name)
        s.add(SystemConstant(name=name, value=value))
        s.commit()
        log.debug(f'Set {name}={value}')
        return value

    @classmethod
    def delete(cls, s, name):
        s.query(SystemConstant).filter(SystemConstant.name == name).delete()
        s.commit()

    TIMEZONE = 'timezone'
    JUPYTER_URL = 'jupyter-url'
    JUPYTER_DIR = 'jupyter-dir'


class SystemProcess(Base):

    __tablename__ = 'system_process'

    id = Column(Integer, primary_key=True)
    command = Column(Text, nullable=True)
    log = Column(Text, nullable=True)
    start = Column(Time, nullable=False, default=time)
    owner = Column(ShortCls, nullable=False, index=True)
    pid = Column(Integer, nullable=False, index=True)

    @classmethod
    def run(cls, s, cmd, log_name, owner):
        process = Popen(args=cmd, shell=True)
        log.debug(f'Adding command [{cmd}] (PID {process.pid})')
        s.add(SystemProcess(command=cmd, owner=owner, pid=process.pid, log=log_name))
        s.commit()
        return process

    @classmethod
    def delete(cls, s, owner, pid, delta_time=3):
        process = s.query(SystemProcess).filter(SystemProcess.owner == owner, SystemProcess.pid == pid).one()
        process.__kill(delta_time=delta_time)
        log.debug(f'Deleting process {process.pid}')
        s.delete(process)
        s.commit()

    @classmethod
    def delete_all(cls, s, owner, delta_time=3):
        for process in s.query(SystemProcess).filter(SystemProcess.owner == owner).all():
            process.__kill(delta_time=delta_time)
            log.debug(f'Deleting process {process.pid}')
            s.delete(process)
        s.commit()

    @classmethod
    def exists_any(cls, s, owner):
        for process in s.query(SystemProcess).filter(SystemProcess.owner == owner).all():
            if process.__still_running():
                return True
        cls.delete_all(s, owner)
        return False

    def __still_running(self, delta_time=3):
        if not pid_exists(self.pid):
            log.debug(f'PID {self.pid} does not exist')
            return False
        if not delta_time:
            return True
        actual = Process(self.pid).create_time()
        saved = self.start.timestamp()
        creation_ok = abs(actual - saved) < delta_time
        if not creation_ok:
            log.debug(f'Creation time for PID {self.pid} incorrect ({actual} / {saved})')
        return creation_ok

    def __kill(self, delta_time=3):
        if self.__still_running(delta_time=delta_time):
            log.debug(f'Killing process {self.pid}')
            Process(self.pid).kill()




from logging import getLogger, basicConfig, INFO
from sys import stdout
from tempfile import NamedTemporaryFile
from unittest import TestCase

from ch2.commands.args import bootstrap_file, m, V
from ch2.lib.data import MutableAttr, reftuple
from ch2.squeal import StatisticJournalFloat, StatisticJournalText, Source
from ch2.squeal.tables.source import SourceType


class TestData(TestCase):

    def setUp(self):
        if not getLogger().handlers:
            basicConfig(stream=stdout, level=INFO)
        self.log = getLogger()

    def test_attr(self):
        d = MutableAttr()
        d['foo'] = 'bar'
        self.assertTrue('foo' in d)
        self.assertEqual(d.foo, 'bar')

    def test_reftuple(self):

        Power = reftuple('Power', 'bike, weight')
        power = Power('#Bike', '$Weight')

        with NamedTemporaryFile() as f:
            args, db = bootstrap_file(f, m(V), '5')
            with db.session_context() as s:
                source = Source(type=SourceType.SOURCE)
                s.add(source)
                StatisticJournalText.add(self.log, s, 'Bike', None, None, self, None, source,
                                         '{"mass": 42}', '1980-01-01')
                StatisticJournalFloat.add(self.log, s, 'Weight', None, None, self, None, source,
                                          13, '1980-01-01')
            p = power.expand(self.log, s, '1990-01-01', owner=self)
            self.assertEqual(p.weight, 13)
            self.assertEqual(p.bike['mass'], 42)

    def test_types(self):
        from ch2.stoats.calculate.power import Power, Bike
        self.assertEqual(Bike.__module__, 'ch2.stoats.calculate.power')
        self.assertEqual(Power.__module__, 'ch2.stoats.calculate.power')




import datetime as dt

from .load import display_notebook
from ..lib.date import format_date, time_to_local_time


def stringify(value):
    if isinstance(value, str):
        return value
    elif isinstance(value, dt.datetime):
        return time_to_local_time(value)
    elif isinstance(value, dt.date):
        return format_date(value)
    else:
        return str(value)


def template(func):

    def wrapper(*args, direct=False, local=True, **kargs):
        if direct:
            return func(*args, **kargs)
        else:
            sargs = [stringify(arg) for arg in args]
            skargs = dict((name, stringify(value)) for name, value in kargs.items())
            display_notebook(func, sargs, skargs, local=local)

    wrapper._original = func

    return wrapper




from abc import abstractmethod
from re import compile

from ...lib.date import format_seconds, to_date, add_date
from ...lib.schedule import Schedule
from ...squeal.tables.topic import Topic
from ...squeal.utils import ORMUtils


class Builder(ORMUtils):

    def __init__(self, name, description, spec, ratio):
        self._name = name
        self._description = description
        self._spec = spec
        self._ratio = ratio

    def create(self, db, parent='Plan', sort=10):
        with db.session_context() as s:
            root = self._get_or_create(s, Topic, name=parent, sort=sort)
            child = Topic(parent=root, schedule=self._spec,
                          name=self._name, description=self._description, sort=sort)
            s.add(child)
            root.schedule = Schedule.include(root.schedule, child.schedule)
            for day in self._spec.locations_from(self._spec.start):
                s.add(Topic(parent=child, schedule=str(day), name=self._next_value()))

    @abstractmethod
    def _next_value(self):
        pass


class TimeBuilder(Builder):

    def __init__(self, name, description, spec, time, ratio):
        self.__time = time
        super().__init__(name, description, spec, ratio)

    def _next_value(self):
        time = self.__time
        self.__time *= self._ratio
        return format_seconds(time)


class DistanceBuilder(Builder):

    def __init__(self, name, description, spec, distance, unit, ratio):
        self.__distance = distance
        self.__unit = unit
        super().__init__(name, description, spec, ratio)

    def _next_value(self):
        distance = self.__distance
        self.__distance *= self._ratio
        return '%.1f%s' % (distance, self.__unit)


def exponential_time(name, repeat, time, percent, start, duration):
    """
    A time interval that increases steadily by a given percentage.
    Takes 6 arguments: name, repeat spec, initial time, percent increase,
                       start date, duration
    """
    time_s = parse_time(time)
    ratio = 1 + float(percent) / 100.0
    spec = Schedule(start + "/" + repeat)
    start = to_date(start)
    finish = add_date(start, parse_duration(duration))
    spec.start = start
    spec.finish = finish
    return TimeBuilder(name, 'Time starting at %s and incrementing by %s%%' % (time, percent),
                       spec, time_s, ratio)


def exponential_distance(name, repeat, distance, percent, start, duration):
    """
    A distance that increases steadily by a given percentage.
    Takes 6 arguments: name, repeat spec, initial distance, percent increase,
                       start date, duration
    """
    match = compile(r'(\d+)(\D*)').match(distance)
    dist, unit = float(match.group(1)), match.group(2)
    ratio = 1 + float(percent) / 100.0
    spec = Schedule(start + "/" + repeat)
    start = to_date(start)
    finish = add_date(start, parse_duration(duration))
    spec.start = start
    spec.finish = finish
    return DistanceBuilder(name, 'Distance starting at %s and incrementing by %s%%' % (distance, percent),
                           spec, dist, unit, ratio)


def parse_time(time):
    '''
    Convert to seconds.  Supports h, m, s.
    '''
    try:
        return int(time)
    except:
        units, time = time[-1].lower(), int(time[:-1])
        return {'s': 1, 'm': 60, 'h': 60*60}[units] * time


def parse_duration(duration):
    '''
    Convert to duration used by lib.date.  Supports d, w, m, y
    '''
    try:
        return int(duration), 'd'
    except:
        units, duration = duration[-1].lower(), duration[:-1]
        if units in 'dwmy':
            return int(duration), units
        else:
            raise




from sqlalchemy import Column, Text, Integer, ForeignKey

from .source import Source, SourceType
from ..types import Time
from ...lib.date import format_time, local_date_to_time


class MonitorJournal(Source):

    __tablename__ = 'monitor_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    fit_file = Column(Text, nullable=False, unique=True)
    start = Column(Time, nullable=False, index=True)
    finish = Column(Time, nullable=False, index=True)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.MONITOR
    }

    def __str__(self):
        return 'Monitor Journal %s to %s' % (format_time(self.start), format_time(self.finish))

    def time_range(self, s):
        return self.start, self.finish






class NullableLog:

    def __init__(self, log):
        self.__log = log

    def set_log(self, log):
        self.__log = log

    def debug(self, *args):
        self.__log.debug(*args)

    def info(self, *args):
        self.__log.info(*args)

    def warning(self, *args):
        self.__log.warning(*args)

    def error(self, *args):
        self.__log.error(*args)


class Named:
    """
    Has a name.  Field for both fields and messages
    """

    def __init__(self, log, name):
        self._log = log
        self.name = name

    def __str__(self):
        return '%s: %s' % (self.__class__.__name__, self.name)


class Rows:

    # next empties __next and then reads from the iter
    # peek gives the future value of next- a preview of next (a single value)
    # multiple lookaheads are possible with lookahead which provides an iterator that parallels
    #   next but appends to __next internally.
    # mixing lookahead and next is not advised!
    # multiple lookaheads in parallel is not advised!

    def __init__(self, sheet, wrapper=tuple):
        self.__rows = (wrapper(cell.value for cell in row) for row in sheet.iter_rows())
        self.__next = []

    def __next__(self):
        if self.__next:
            value, self.__next = self.__next[0], self.__next[1:]
        else:
            value = next(self.__rows)
        return value

    def __iter__(self):
        return self

    def __bool__(self):
        return bool(self.peek())

    def peek(self):
        if not self.__next:
            try:
                self.__next.append(next(self.__rows))
            except StopIteration:
                pass
        if self.__next:
            return self.__next[0]
        else:
            return None

    def lookahead(self):
        yield from self.__next
        try:
            while True:
                self.__next.append(next(self.__rows))
                yield self.__next[-1]
        except StopIteration:
            return







from contextlib import contextmanager
from logging import getLogger
from sqlite3 import OperationalError

from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.functions import count

from . import *
from .support import Base
from ..commands.args import DATABASE, NamespaceWithVariables, NO_OP, parser
from ..lib.log import make_log

# mention these so they are "created" (todo - is this needed? missing tables seem to get created anyway)
Source,  Interval, Dummy, Composite, CompositeComponent
ActivityGroup, ActivityJournal, ActivityTimespan, ActivityBookmark
Topic, TopicJournal, TopicField,
StatisticName, StatisticJournal, StatisticJournalInteger, StatisticJournalFloat, StatisticJournalText, StatisticMeasure
Segment, SegmentJournal
Pipeline
MonitorJournal
Constant, SystemConstant, SystemProcess
ActivitySimilarity, ActivityNearby
Timestamp

log = getLogger(__name__)

# https://stackoverflow.com/questions/13712381/how-to-turn-on-pragma-foreign-keys-on-in-sqlalchemy-migration-script-or-conf
@event.listens_for(Engine, "connect")
def fk_pragma_on_connect(dbapi_con, _con_record):
    cursor = dbapi_con.cursor()

    def pragma(cmd):
        full_cmd = f'PRAGMA {cmd}'
        # log.debug(full_cmd)
        cursor.execute(full_cmd)

    pragma('foreign_keys=ON;')  # https://www.sqlite.org/pragma.html#pragma_foreign_keys
    pragma('temp_store=MEMORY;')  # https://www.sqlite.org/pragma.html#pragma_temp_store
    pragma('threads=4;')  # https://www.sqlite.org/pragma.html#pragma_threads
    pragma('cache_size=-1000000;')  # 1GB  https://www.sqlite.org/pragma.html#pragma_cache_size
    pragma('secure_delete=OFF;')  # https://www.sqlite.org/pragma.html#pragma_secure_delete
    pragma('journal_mode=WAL;')  # https://www.sqlite.org/wal.html
    pragma(f'busy_timeout={5 * 60 * 1000};')  # https://www.sqlite.org/pragma.html#pragma_busy_timeout
    cursor.close()


@event.listens_for(Engine, 'close')
def analyze_pragma_on_close(dbapi_con, _con_record):
    cursor = dbapi_con.cursor()
    try:
        # this can fail if another process is using the database
        # log.debug('Optimize DB...')
        cursor.execute("PRAGMA optimize;")  # https://www.sqlite.org/pragma.html#pragma_optimize
        # log.debug('Optimize DB done')
    except OperationalError as e:
        log.debug("Optimize DB aborted (DB Likely still in use)")
    finally:
        cursor.close()


class Database:

    def __init__(self, args, log):
        self._log = log
        path = args.file(DATABASE)
        self._log.info('Using database at %s' % path)
        self.engine = create_engine('sqlite:///%s' % path, echo=False)
        self.session = sessionmaker(bind=self.engine)
        self.__create_tables()

    def __create_tables(self):
        if self.is_empty(tables=True):
            self._log.info('Creating tables')
            Base.metadata.create_all(self.engine)

    @contextmanager
    def session_context(self):
        session = self.session()
        try:
            yield session
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def is_empty(self, tables=False):
        if tables:
            # https://stackoverflow.com/questions/33053241/sqlalchemy-if-table-does-not-exist
            return not self.engine.dialect.has_table(self.engine, Source.__tablename__)
        else:
            with self.session_context() as s:
                n_topics = s.query(count(Topic.id)).scalar()
                n_activities = s.query(count(ActivityGroup.id)).scalar()
                n_statistics = s.query(count(StatisticName.id)).scalar()
            return not (n_topics + n_activities + n_statistics)


def connect(args):
    '''
    Bootstrap from commandline-like args.
    '''
    if len(args) == 1:
        args = args[0].split()
    elif args:
        args = list(args)
    else:
        args = []
    args.append(NO_OP)
    ns = NamespaceWithVariables(parser().parse_args(args))
    make_log(ns)
    db = Database(ns, log)
    return ns, db





import datetime as dt
from abc import abstractmethod
from collections import namedtuple
from re import compile
from struct import unpack, pack

from .support import Named, Rows
from ...lib.data import WarnDict, WarnList

LITTLE, BIG = 0, 1


class AbstractType(Named):
    '''
    Root class for any kind of type in the system.
    '''

    def __init__(self, log, name, n_bytes, base_type=None):
        super().__init__(log, name)
        self.base_type = base_type
        self.n_bytes = n_bytes

    def is_bad(self, bytes, count, endian):
        return False

    @abstractmethod
    def profile_to_internal(self, cell_contents):
        raise NotImplementedError('%s: %s' % (self.__class__.__name__, self.name))

    @abstractmethod
    def parse_type(self, bytes, count, endian, timestamp, **options):
        raise NotImplementedError('%s: %s' % (self.__class__.__name__, self.name))


class SimpleType(AbstractType):
    '''
    Value is extracted directly from the binary data.
    '''

    def __init__(self, log, name, n_bytes, func):
        super().__init__(log, name, n_bytes)
        self.__func = func

    def profile_to_internal(self, cell_contents):
        return self.__func(cell_contents)


class StructSupport(SimpleType):
    '''
    Most base types use stucts to unpack data.
    '''

    def _pack_bad(self, value):
        bad = (bytearray(self.n_bytes), bytearray(self.n_bytes))
        for endian in (LITTLE, BIG):
            bytes = value
            for i in range(self.n_bytes):
                j = i if endian == LITTLE else self.n_bytes - i - 1
                bad[endian][j] = bytes & 0xff
                bytes >>= 8
        return bad

    def _all_bad(self, data, bad, count):
        return all(bad == data[self.n_bytes*i:self.n_bytes*(i+1)] for i in range(count))

    # currently this ignores scale and offset!!!
    def _pack(self, values, formats, count, endian):
        return pack(formats[endian] % count, *values)

    # scale and offset have to be at this level because of how bad values when count > 1 are handled
    def _unpack(self, data, formats, bad, count, endian, scale=1, offset=0, check_bad=True,
                name=None, accumulators=None, n_bits=None, **options):
        if check_bad and self._all_bad(data, bad[endian], count):
            return None
        elif accumulators and name in accumulators:
            if count > 1:
                raise Exception('Cannot accumulate multiple fields (%s: %d)' % (name, count))
            return self.__unpack_acc(bytearray(data[:self.n_bytes]), formats[endian] % 1, scale, offset,
                                     name, accumulators, n_bits, endian)
        else:
            if (scale == 1 and offset == 0) or self.name == 'enum':   # enums are not scaled
                # fast and preserves integers
                return unpack(formats[endian] % count, data[:self.n_bytes * count])
            elif count == 1:  # if no check, scale single bad values
                return (unpack(formats[endian] % 1, data[:self.n_bytes])[0] / scale - offset,)
            else:  # match weird CSV behaviour
                return tuple(self.__unpack_scaled(data[self.n_bytes*i:self.n_bytes*(i+1)], formats[endian],
                                                  bad[endian], scale, offset) for i in range(count))

    def __unpack_scaled(self, data, format, bad, scale, offset):
        value = unpack(format % 1, data)[0]
        if data == bad or (scale == 1 and offset == 0):
            # the java CSV program returns isolated bad values in multiples as unscaled
            return value
        else:
            return value / scale - offset

    def __unpack_acc(self, data, format, scale, offset, name, accumulators, n_bits, endian):
        short = unpack(format, data)[0]
        if accumulators[name] is not None:
            prev_data, prev_long = accumulators[name]
            if n_bits:
                mask = (1 << n_bits) - 1
                prev_short = prev_long & mask
            elif short < prev_long:
                raise Exception('Full accumulated field has decreased in value (%s: %d/%d)' %
                                (name, short, prev_long))  # or n_bits was missing...
            if short >= prev_long:  # initial phase of simple growth OR a full read
                long = short
            else:  # otherwise, need all bits
                data = self.__merge_bytes(prev_data, data, n_bits, endian)
                if short < prev_short:  # rollover?
                    data = self.__add_bit(data, n_bits, endian)
                long = unpack(format, data)[0]
        else:
            long = short
        accumulators[name] = (data, long)
        if scale == 1 and offset == 0:
            return (long,)
        else:
            return (long / scale - offset,)

    def __merge_bytes(self, prev_data, data, n_bits, endian, index=None):
        now = min(8, n_bits)
        later = max(0, n_bits - now)
        if endian == LITTLE:
            if index is None: index = 0
            prev_data[index] = self.__merge_bits(prev_data[index], data[index], now)
            if later:
                self.__merge_bytes(prev_data, data, later, endian, index=index+1)
        else:
            if index is None: index = -1
            prev_data[index] = self.__merge_bits(prev_data[index], data[index], now)
            if later:
                self.__merge_bytes(prev_data, data, later, endian, index=index-1)
        return prev_data

    def __merge_bits(self, prev_data, data, n_bits):
        mask = (1 << n_bits) - 1
        return (prev_data & ~mask) | (data & mask)

    def __add_bit(self, data, n_bits, endian):
        value = 1 << (n_bits % 8)
        index = n_bits // 8
        if endian == LITTLE:
            while value and index < len(data):
                value += data[index]
                data[index] = value & 0xff
                value = (value & 0x100) >> 8
                index += 1
        else:
            while value and index >= 0:
                value += data[index]
                data[index] = value & 0xff
                value = (value & 0x100) >> 8
                index -= 1
        return data


class String(SimpleType):

    def __init__(self, log, name):
        super().__init__(log, name, 1, str)

    def parse_type(self, bytes, count, endian, timestamp, **options):
        try:
            value = str(b''.join(unpack('%dc' % count, bytes)), encoding='utf-8')
            while value and value[-1] == '\u0000': value = value[:-1]
            return value.split('\u0000')  # inferred from CSV handling of weird data in tests
        except UnicodeDecodeError:
            # this gets us close to what garmin generates in CSV files
            bytes = bytearray(b % 0x7f for b in bytes)
            while bytes and bytes[-1] == 0: bytes = bytes[:-1]
            return tuple(b.decode('ascii') for b in bytes.split(b'\0'))


class Boolean(SimpleType):

    def __init__(self, log, name):
        super().__init__(log, name, 1, bool)

    def parse_type(self, bytes, count, endian, timestamp, **options):
        return tuple(bool(byte) for byte in bytes)


class AutoInteger(StructSupport):
    '''
    A whole pile of different integer types can all be parameterised by sign and size.
    '''

    pattern = compile(r'^([su]?)int(\d{1,2})(z?)$')

    size_to_format = {1: 'b', 2: 'h', 4: 'i', 8: 'q'}

    def __init__(self, log, name):
        match = self.pattern.match(name)
        self.signed = match.group(1) != 'u'
        n_bits = int(match.group(2))
        if n_bits % 8:
            raise Exception('Size of %r not a multiple of 8 bits' % name)
        bad = 0 if match.group(3) == 'z' else 1
        super().__init__(log, name, n_bits // 8, self.int)
        if self.n_bytes not in self.size_to_format:
            raise Exception('Cannot unpack %d bytes as an integer' % self.n_bytes)
        format = self.size_to_format[self.n_bytes]
        if not self.signed:
            format = format.upper()
        self.__formats = ['<%d' + format, '>%d' + format]
        self.__bad = self._pack_bad(0 if match.group(3) == 'z' else 2 ** (n_bits - (1 if self.signed else 0)) - 1)

    @staticmethod
    def int(cell):
        if isinstance(cell, int):
            return cell
        else:
            return int(cell, 0)

    def is_bad(self, bytes, count, endian):
        return self._all_bad(bytes, self.__bad[endian], count)

    def parse_type(self, data, count, endian, timestamp, check_bad=True, **options):
        return self._unpack(data, self.__formats, self.__bad, count, endian, check_bad=check_bad, **options)

    def pack_type(self, values, count, endian):
        return self._pack(values, self.__formats, count, endian)


class AliasInteger(AutoInteger):
    '''
    An integer with an alternative name.
    '''

    def __init__(self, log, name, spec):
        super().__init__(log, spec)
        self.name = name


def timestamp_to_time(timestamp, tzinfo=dt.timezone.utc):
    return dt.datetime(1989, 12, 31, tzinfo=tzinfo) + dt.timedelta(seconds=timestamp)


def time_to_timestamp(time, tzinfo=dt.timezone.utc):
    return int((time - dt.datetime(1989, 12, 31, tzinfo=tzinfo)).total_seconds())


class BadTimestamp(Exception): pass


class Date(AliasInteger):

    def __init__(self, log, name, utc=True):
        super().__init__(log, name, 'uint32')
        self.__tzinfo = dt.timezone.utc if utc else None

    def convert(self, time, tzinfo=dt.timezone.utc):
        if time is not None:
            return timestamp_to_time(time, tzinfo=tzinfo)

    def parse_type(self, data, count, endian, timestamp, raw_time=False, **options):
        times = super().parse_type(data, count, endian, timestamp, raw_time=raw_time, **options)
        if times and not raw_time:
            times = tuple(self.convert(time, tzinfo=self.__tzinfo) for time in times)
        return times

    def pack_type(self, values, count, endian):
        return super().pack_type([time_to_timestamp(value) for value in values], count, endian)


class Date16(AliasInteger):

    def __init__(self, log, name, utc=True):
        super().__init__(log, name, 'uint16')
        self.__tzinfo = dt.timezone.utc if utc else None

    def convert(self, time, timestamp, tzinfo=dt.timezone.utc):
        # https://www.thisisant.com/forum/viewthread/6374
        current = time_to_timestamp(timestamp, tzinfo=tzinfo)
        current += (time - (current & 0xffff)) & 0xffff
        return timestamp_to_time(current, tzinfo=tzinfo)

    def parse_type(self, data, count, endian, timestamp, raw_time=False, **options):
        times = super().parse_type(data, count, endian, timestamp, raw_time=raw_time, **options)
        if not raw_time:
            times = tuple(self.convert(time, timestamp, tzinfo=self.__tzinfo) for time in times)
        return times


class AutoFloat(StructSupport):

    pattern = compile(r'^float(\d{1,2})$')

    size_to_format = {2: 'e', 4: 'f', 8: 'd'}

    def __init__(self, log, name):
        match = self.pattern.match(name)
        n_bits = int(match.group(1))
        if n_bits % 8:
            raise Exception('Size of %r not a multiple of 8 bits' % name)
        super().__init__(log, name, n_bits // 8, float)
        if self.n_bytes not in self.size_to_format:
            raise Exception('Cannot unpack %d bytes as a float' % self.n_bytes)
        format = self.size_to_format[self.n_bytes]
        self.__formats = ['<%d' + format, '>%d' + format]
        self.__bad = self._pack_bad(2 ** n_bits - 1)

    def is_bad(self, bytes, count, endian):
        return self._all_bad(bytes, self.__bad[endian], count)

    def parse_type(self, data, count, endian, timestamp, check_bad=True, **options):
        return self._unpack(data, self.__formats, self.__bad, count, endian, check_bad=check_bad, **options)


class Mapping(AbstractType):

    def __init__(self, log, row, rows, types, warn=False):
        name = row.type_name
        base_type_name = row.base_type
        base_type = types.profile_to_type(base_type_name, auto_create=True)
        super().__init__(log, name, base_type.n_bytes, base_type=base_type)
        self._profile_to_internal = WarnDict(log, 'No internal value for profile %r') if warn else dict()
        self._internal_to_profile = WarnDict(log, 'No profile value for internal %r') if warn else dict()
        while rows:
            peek = rows.peek()
            if peek.type_name or peek.value_name is None or peek.value is None:
                return
            self.__add_mapping(next(rows))

    def profile_to_internal(self, cell_contents):
        return self._profile_to_internal[cell_contents]

    def internal_to_profile(self, value):
        return self._internal_to_profile[value]

    def safe_internal_to_profile(self, value):
        try:
            return self.internal_to_profile(value)
        except KeyError:
            return value

    # default here for check_bad sets whether mappings can be considered bad or not
    # tests against CSV suggest they can (battery_level)
    def parse_type(self, bytes, size, endian, timestamp, map_values=True, check_bad=True, **options):
        values = self.base_type.parse_type(bytes, size, endian, timestamp, check_bad=check_bad, **options)
        if map_values and values:
            values = tuple(self.safe_internal_to_profile(value) for value in values)
        return values

    def __add_mapping(self, row):
        profile = row.value_name
        internal = self.base_type.profile_to_internal(row.value)
        self._profile_to_internal[profile] = internal
        self._internal_to_profile[internal] = profile


BASE_TYPE_NAMES = ['enum', 'sint8', 'uint8', 'sint16', 'uint16', 'sint32', 'uint32',
                   'string', 'float32', 'float64',
                   'uint8z', 'uint16z', 'uint32z', 'byte', 'sint64', 'uint64', 'uint64z']


class Types:

    def __init__(self, log, sheet, warn=False):
        self.__log = log
        self.__profile_to_type = WarnDict(log, 'No type for profile %r')
        # these are not 'base types' in the same sense as types having base types.
        # rather, they are the 'base (integer) types' described in the docs
        self.base_types = WarnList(log, 'No base type for number %r')
        self.overrides = set()
        self.__add_known_types()
        rows = Rows(sheet, wrapper=Row)
        for row in rows:
            if row[0] and row.type_name[0].isupper():
                self.__log.debug('Skipping %s' % (row,))
            elif row[0]:
                # self.__log.info('Parsing type %s' % row[0])
                self.__add_type(Mapping(self.__log, row, rows, self, warn=warn))

    def __add_known_types(self):
        # these cannot be inferred from name
        self.__add_type(String(self.__log, 'string'))
        self.__add_type(AliasInteger(self.__log, 'enum', 'uint8'))
        self.__add_type(AliasInteger(self.__log, 'byte', 'uint8'))
        # these can be inferred
        for name in BASE_TYPE_NAMES:
            self.profile_to_type(name, auto_create=True)
            self.base_types.append(self.profile_to_type(name))
        # this is in the spreadsheet, but not in the doc
        self.__add_type(Boolean(self.__log, 'bool'))
        # these are defined in the spreadsheet, but the interpretation is in comments
        self.__add_override(Date(self.__log, 'date_time', utc=True))
        self.__add_override(Date(self.__log, 'local_date_time', utc=False))
        self.__add_override(Date16(self.__log, 'timestamp_16', utc=True))

    def __add_override(self, type):
        self.overrides.add(type.name)
        self.__add_type(type)

    def __add_type(self, type):
        if type.name in self.__profile_to_type:
            duplicate = self.__profile_to_type[type.name]
            if duplicate.n_bytes == type.n_bytes:
                self.__log.warning('Ignoring duplicate type for %r' % type.name)
            else:
                raise Exception('Duplicate type for %r with differing size (%d  %d)' %
                                (type.name, type.n_bytes, duplicate.n_bytes))
        else:
            self.__profile_to_type[type.name] = type

    def is_type(self, name):
        return name in self.__profile_to_type

    def profile_to_type(self, name, auto_create=False):
        try:
            return self.__profile_to_type[name]
        except KeyError:
            if auto_create:
                for cls in (AutoFloat, AutoInteger):
                    match = cls.pattern.match(name)
                    if match:
                        self.__log.info('Auto-adding type %s for %r' % (cls.__name__, name))
                        self.__add_type(cls(self.__log, name))
                        return self.profile_to_type(name)
            raise


class Row(namedtuple('BaseRow',
                     'type_name, base_type, value_name, value, comment')):

    __slots__ = ()

    def __new__(cls, row):
        return super().__new__(cls, *tuple(row)[0:5])





import datetime as dt
from math import sqrt

import pandas as pd
from bokeh import palettes, tile_providers
from bokeh.layouts import column, row
from bokeh.models import PanTool, ZoomInTool, ZoomOutTool, ResetTool, HoverTool, Range1d, LinearAxis
from bokeh.plotting import figure

from .frame import present
from ..stoats.names import LOCAL_TIME, TIMESPAN_ID, TIME, CLIMB_DISTANCE, CLIMB_ELEVATION, SPHERICAL_MERCATOR_X, \
    SPHERICAL_MERCATOR_Y, LATITUDE, LONGITUDE, DISTANCE_KM, ELEVATION_M


def col_to_boxstats(frame, name):
    '''
    Allow inter-op with matplotlib using pre-calculated stats
    See test_data.py
    '''
    stats = []
    for index, row in frame.iterrows():
        if row[name]:
            stats.append({'label': index,
                          'whislo': row[name][0],
                          'q1': row[name][1],
                          'med': row[name][2],
                          'q3': row[name][3],
                          'whishi': row[name][4]})
    return stats


def box_plot(f, col):
    '''
    Generate a boxplot for a column (pandas series) containing a tuple of 5 values
    (index date) as provided by summary statistics,
    '''
    def pick(n):
        def pick(x):
            return x[n] if x else None
        return pick
    q = [col.map(pick(i)) for i in range(5)]
    f.segment(q[0].index, q[0], q[1].index, q[1])
    f.vbar(q[1].index, dt.timedelta(days=20), q[1], q[2], fill_alpha=0)
    f.vbar(q[2].index, dt.timedelta(days=20), q[2], q[3], fill_alpha=0)
    f.segment(q[3].index, q[3], q[4].index, q[4])


def tooltip(name):
    return name, '@{%s}' % name


def make_tools(y):
    tools = [PanTool(dimensions='width'),
             ZoomInTool(dimensions='width'), ZoomOutTool(dimensions='width'),
             ResetTool(),
             HoverTool(tooltips=[tooltip(x) for x in (y, DISTANCE_KM, LOCAL_TIME)])]
    return tools


def subtract(a, c, key, col):
    cols = [key, col]
    limit = min(a[key].max(), c[key].max())
    a = a.loc[a[key] <= limit, cols]
    c = c.loc[c[key] <= limit, cols]
    both = a.merge(c, how='outer', on=key, sort=True, suffixes=('_a', '_c'))
    both.interpolate(inplace=True, limit_direction='both')
    return pd.DataFrame({col: both[col+'_a'] - both[col+'_c'], key: both[key]})


def patches(x, y, diff):
    y = diff.set_index(x)[y]
    range = y.abs().max() * 1.1
    green = y.clip(lower=0).append(pd.Series([0, 0], index=[y.index[len(y)-1], y.index[0]]))
    red = y.clip(upper=0).append(pd.Series([0, 0], index=[y.index[len(y)-1], y.index[0]]))
    return green, red, Range1d(start=-range, end=range)


def y_range(f, y, source, ylo=None, yhi=None):
    ylo, yhi = source[y].dropna().min() if ylo is None else ylo, source[y].dropna().max() if yhi is None else yhi
    dy = yhi - ylo
    f.y_range = Range1d(start=ylo - 0.1 * dy, end=yhi + 0.1 * dy)


def add_tsid_line(f, x, y, source, color='black', line_dash='solid'):
    for _, s in source.groupby(TIMESPAN_ID):
        f.line(x=x, y=y, source=s, line_color=color, line_dash=line_dash)


def comparison_line_plot(nx, ny, x, y, source, other=None, ylo=None, yhi=None, x_range=None):
    f = figure(plot_width=nx, plot_height=ny, x_axis_type='datetime' if TIME in x else 'linear', tools=make_tools(y))
    y_range(f, y, source, ylo=ylo, yhi=yhi)
    add_tsid_line(f, x, y, source)
    if present(other, y):
        add_tsid_line(f, x, y, other, color='grey')
        diff = subtract(source, other, x, y)
        green, red, y_range2 = patches(x, y, diff)
        f.extra_y_ranges = {'delta': y_range2}
        f.patch(x=green.index, y=green, color='green', alpha=0.1, y_range_name='delta')
        f.patch(x=red.index, y=red, color='red', alpha=0.1, y_range_name='delta')
    f.yaxis.axis_label = y
    f.xaxis.axis_label = x
    f.toolbar.logo = None
    if x_range: f.x_range = x_range
    return f


def add_cum_line(f, y, source, color='black', line_dash='solid'):
    y_c = source[y].sort_values(ascending=False).reset_index(drop=True)
    f.line(x=y_c.index, y=y_c, line_color=color, line_dash=line_dash)
    f.x_range = Range1d(start=y_c.index.max(), end=y_c.index.min())
    df = y_c.to_frame('y')
    df['x'] = df.index
    return df


def cumulative_plot(nx, ny, y, source, other=None, ylo=None, yhi=None):
    f = figure(plot_width=nx, plot_height=ny, y_axis_location='right')
    y_range(f, y, source, ylo=ylo, yhi=yhi)
    y1 = add_cum_line(f, y, source)
    if present(other, y):
        y2 = add_cum_line(f, y, other, color='grey')
        diff = subtract(y1, y2, 'x', 'y')
        green, red, y_range2 = patches('x', 'y', diff)
        f.extra_y_ranges = {'delta': y_range2}
        f.patch(x=green.index, y=green, color='green', alpha=0.1, y_range_name='delta')
        f.patch(x=red.index, y=red, color='red', alpha=0.1, y_range_name='delta')
    f.xaxis.visible = False
    f.yaxis.axis_label = y
    f.toolbar_location = None
    return f


def add_climbs(f, climbs, source):
    for time, climb in climbs.loc[~pd.isna(climbs[CLIMB_DISTANCE])].iterrows():
        i = source.index.get_loc(time, method='nearest')
        x = source[DISTANCE_KM].iloc[i]
        x = (x - climb[CLIMB_DISTANCE] / 1000, x)
        y = source[ELEVATION_M].iloc[i]
        y = (y - climb[CLIMB_ELEVATION], y)
        f.line(x=x, y=y, color='red', line_width=5, alpha=0.2)
        for xx, yy in zip(x, y):
            f.circle(x=xx, y=yy, color='red', size=8, alpha=0.2)


def histogram_plot(nx, ny, x, source, xlo=None, xhi=None, nsub=5):
    xlo, xhi = source[x].min() if xlo is None else xlo, source[x].max() if xhi is None else xhi
    bins = pd.interval_range(start=xlo, end=xhi, periods=nsub * (xhi - xlo), closed='left')
    c = [palettes.Inferno[int(xhi-xlo+1)][int(b.left-xlo)] for b in bins]
    hrz_categorized = pd.cut(source[x], bins)
    counts = hrz_categorized.groupby(hrz_categorized).count()
    f = figure(plot_width=nx, plot_height=ny, x_range=Range1d(start=xlo, end=xhi), x_axis_label=x)
    f.quad(left=counts.index.categories.left, right=counts.index.categories.right, top=counts, bottom=0,
           color=c, fill_alpha=0.2)
    f.toolbar_location = None
    f.yaxis.visible = False
    return f


def add_route(f, source, color='black', line_dash='solid'):
    return f.line(x=SPHERICAL_MERCATOR_X, y=SPHERICAL_MERCATOR_Y, source=source,
                  color=color, line_dash=line_dash)


def map_plot(nx, ny, source, other=None):
    tools = [PanTool(dimensions='both'),
             ZoomInTool(dimensions='both'), ZoomOutTool(dimensions='both'),
             ResetTool(),
             HoverTool(tooltips=[tooltip(x) for x in (LATITUDE, LONGITUDE, DISTANCE_KM, LOCAL_TIME)])]
    f = figure(plot_width=nx, plot_height=ny, x_axis_type='mercator', y_axis_type='mercator', tools=tools)
    add_route(f, source)
    if present(other, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y):
        add_route(f, other, color='black', line_dash='dotted')
    f.add_tile(tile_providers.STAMEN_TERRAIN, alpha=0.3)
    f.axis.visible = False
    f.toolbar.logo = None
    return f


def map_intensity(nx, ny, source, z, power=1.0, color='red', alpha=0.01, ranges=None):
    tools = [PanTool(dimensions='both'),
             ZoomInTool(dimensions='both'), ZoomOutTool(dimensions='both'),
             ResetTool(),
             HoverTool(tooltips=[tooltip(x) for x in (z, DISTANCE_KM, LOCAL_TIME)])]
    f = figure(plot_width=nx, plot_height=ny, x_axis_type='mercator', y_axis_type='mercator',
               title=z, tools=tools)
    tools[-1].renderers = [add_route(f, source)]
    mn, mx = source[z].min(), source[z].max()
    source['size'] = sqrt(nx * ny) * ((source[z] - mn) / (mx - mn)) ** power / 10
    f.circle(x=SPHERICAL_MERCATOR_X, y=SPHERICAL_MERCATOR_Y, size='size', source=source, color=color, alpha=alpha)
    f.axis.visible = False
    f.toolbar.logo = None
    if ranges is not None:
        f.x_range = ranges.x_range
        f.y_range = ranges.y_range
    return f


def map_thumbnail(nx, ny, source, other=None):
    f = figure(plot_width=nx, plot_height=ny, x_axis_type='mercator', y_axis_type='mercator',
               title=source.index[0].strftime('%Y-%m-%d'))
    add_route(f, source)
    f.axis.visible = False
    f.toolbar_location = None
    return f


def line_plotter():
    return lambda f, *args, **kargs: f.line(*args, **kargs)


def dot_plotter():
    return lambda f, *args, **kargs: f.circle(*args, **kargs)


def bar_plotter(delta):
    def plotter(f, x=None, y=None, source=None, **kargs):
        f.vbar(x=x, width=delta, top=y, source=source, **kargs)
    return plotter


def multi_line_plot(nx, ny, x, ys, source, colors, alphas=None, x_range=None, y_label=None, rescale=False):
    return multi_plot(nx, ny, x, ys, source, colors, alphas=alphas, x_range=x_range, y_label=y_label, rescale=rescale,
                      plotters=[line_plotter()])


def multi_dot_plot(nx, ny, x, ys, source, colors, alphas=None, x_range=None, y_label=None, rescale=False):
    return multi_plot(nx, ny, x, ys, source, colors, alphas=alphas, x_range=x_range, y_label=y_label, rescale=rescale,
                      plotters=[dot_plotter()])


def multi_bar_plot(nx, ny, x, ys, source, colors, alphas=None, x_range=None, y_label=None, rescale=False):
    return multi_plot(nx, ny, x, ys, source, colors, alphas=alphas, x_range=x_range, y_label=y_label, rescale=rescale,
                      plotters=[bar_plotter(dt.timedelta(hours=20))])


def multi_plot(nx, ny, x, ys, source, colors, alphas=None, x_range=None, y_label=None, rescale=False,
               plotters=None):
    tools = [PanTool(dimensions='width'),
             ZoomInTool(dimensions='width'), ZoomOutTool(dimensions='width'),
             ResetTool(),
             HoverTool(tooltips=[tooltip(x) for x in ys + [LOCAL_TIME]])]
    f = figure(plot_width=nx, plot_height=ny, x_axis_type='datetime' if TIME in x else 'linear', tools=tools)
    if y_label:
        f.yaxis.axis_label = y_label
    elif rescale:
        f.yaxis.axis_label = ys[0]
    else:
        f.yaxis.axis_label = ', '.join(ys)
    if rescale: f.extra_y_ranges = {}
    if alphas is None: alphas = [1 for _ in ys]
    while len(plotters) < len(ys): plotters += plotters
    for y, color, alpha, plotter in zip(ys, colors, alphas, plotters):
        mn, mx = source[y].dropna().min(), source[y].dropna().max()
        dy = mx - mn
        if rescale and y != ys[0]:
            f.extra_y_ranges[y] = Range1d(start=mn - 0.1 * dy, end=mx + 0.1 * dy)
            f.add_layout(LinearAxis(y_range_name=y, axis_label=y), 'right')
            plotter(f, x=x, y=y, source=source, color=color, alpha=alpha, y_range_name=y)
        else:
            f.y_range = Range1d(start=mn - 0.1 * dy, end=mx + 0.1 * dy)
            plotter(f, x=x, y=y, source=source, color=color, alpha=alpha)
    f.xaxis.axis_label = x
    f.toolbar.logo = None
    if ny < 300: f.toolbar_location = None
    if x_range: f.x_range = x_range
    return f


def tile(maps, n):
    return column([row(maps[i:i+n]) for i in range(0, len(maps), n)])